/* manette-mapping-error-private.h
 *
 * Copyright (C) 2017 Adrien Plazas <kekun.plazas@laposte.net>
 *
 * This file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the
 * License, or (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MANETTE_MAPPING_ERROR_H
#define MANETTE_MAPPING_ERROR_H

#if !defined(__MANETTE_INSIDE__) && !defined(MANETTE_COMPILATION)
# error "Only <libmanette.h> can be included directly."
#endif

#include <glib.h>

G_BEGIN_DECLS

#define MANETTE_MAPPING_ERROR (manette_mapping_error_quark ())

typedef enum
{
  MANETTE_MAPPING_ERROR_NOT_A_MAPPING,
} ManetteMappingError;

GQuark manette_mapping_error_quark (void);

G_END_DECLS

#endif /* MANETTE_MAPPING_ERROR_H */
